#' Based on true results and extreme variants, add protection error and report approximate values and approximate protection error.
#' @param .g list of temporary objects needed.
#' @return Updated version of .g.
.jk.univariate.finish <- function(.g){
	# requires:
	#	mysql.connection
	#	true.and.extreme.results
	#	stat.labels
	#   stat.names
	# assigns:
	

	# handy abbreviations:

	x <- .g$analyse.names
	g <- .g$group.by.ids
	g.sql <- .g$group.by.ids.sql
	g.t.sql <- cpaste("t.",g)
		

	# prepare db:

	dbSendQuery(.g$mysql.connection, "DROP TABLE IF EXISTS _secure_results0" )
	dbSendQuery(.g$mysql.connection, "DROP TABLE IF EXISTS _secure_results" )


	# prepare column lists:

	s0 <- rep(.g$stat.names,rep(length(x),length(.g$stat.names)))
	x0 <- rep(x,length(.g$stat.names))


	# compute maximum deviations:
	
	max.deviations.view <- cpaste(
	
		 " ( SELECT t.*, "
			,cpaste(
				"  GREATEST(ABS(u_",s0,"__",x0,"-",s0,"__",x0,"),ABS(",s0,"__",x0,"-l_",s0,"__",x0,")) AS d_",s0,"__",x0
			)
		," FROM ",.g$true.and.extreme.results," AS t "
        ," GROUP BY ",cpaste("t.",g)
		," ) "
	)
	

	# compute unique hashs for each combination of statistics*column*group:

	hashs.view <- cpaste(
	
		 " ( SELECT ",g.sql,", "
			,cpaste(
				" MD5(CONCAT('",s0,"',sb__",x0,")) AS hash_",s0,"__",x0
			)
		," FROM ",.g$seed.base," AS s ) "
	)
	
	if (.g$.jk.debug) {
		print("max.deviations.view")
		print(max.deviations.view)
		print(dbGetQuery(.g$mysql.connection,max.deviations.view))	
		print("hashs.view")
		print(hashs.view)
		print(dbGetQuery(.g$mysql.connection,hashs.view))	
	}	
	
	
	# compute secure results for non-empty groups
	# by adding a Normal(0,d_...)-distributed protection error
	# and include a Weibull(4,1.37*d_...)-distributed estimate 
	# of the protection error's dispersion ("perr_...")
	# (gives a 75.89% conf. bound on the true value, its double a 95.77% conf. bound, its triple a 99% conf. bound)
	
	dbSendQuery(.g$mysql.connection,cpaste(
	
		 " CREATE TABLE _secure_results0 "
		," AS SELECT ",g.t.sql,", "
			,cpaste(
				 s0,"__",x0," + d_",s0,"__",x0
				," * SQRT(22.180709777918-LOG( CONV(SUBSTR(hash_",s0,"__",x0,",1,8),16,10) )) "
									 ," * COS( CONV(SUBSTR(hash_",s0,"__",x0,",9,8),16,10) ) AS ",s0,"__",x0
			), ", "
			,cpaste(
				 "d_",s0,"__",x0
				," * 1.37 * SQRT(SQRT(22.180709777918-LOG( CONV(SUBSTR(hash_",s0,"__",x0,",17,8),16,10) ))) "
				," AS perr_",s0,"__",x0
			)
			,ifelse(.g$.jk.debug,cpaste(", "
				,cpaste(
				 	s0,"__",x0," AS true_",s0,"__",x0
					," ,d_",s0,"__",x0
				), ", "
				,cpaste(
				 	"d_",s0,"__",x0,"/ABS(",s0,"__",x0,") AS rel_",s0,"__",x0
				)
			),"")
		," FROM ",hashs.view," AS h, ",max.deviations.view," AS t"
		," WHERE "
			,paste(
				"t.",g,"=h.",g
				,sep="",collapse=" AND ")
		," GROUP BY ",cpaste("t.",g)
	))

	if (.g$.jk.debug) {
		print("_secure_results0")
		print(dbReadTable(.g$mysql.connection,"_secure_results0"))
	}


	# add rows for empty groups:
	
	dbSendQuery(.g$mysql.connection, cpaste(
	
		 " CREATE TABLE _secure_results "
		," AS SELECT ",cpaste("t.",g),", "
			,ifelse(.g$.jk.debug,cpaste(
				 s0,"__",x0
				,", perr_",s0,"__",x0
				,", ABS( ",s0,"__",x0,"-true_",s0,"__",x0," ) AS err_",s0,"__",x0
			),cpaste(
				 s0,"__",x0
				,", perr_",s0,"__",x0
			))	
		," FROM _secure_results0 AS s "
		," RIGHT JOIN _groups AS t "
		," ON "
			,paste(
				"s.",g,"=t.",g
				,sep="",collapse=" AND ")
		," GROUP BY ",cpaste("t.",g)
	))
    secure.results <- dbReadTable(.g$mysql.connection,"_secure_results")
	if (.g$.jk.debug) {
		print("_secure_results")
		print(secure.results)
	}
    
	# get a stream in the order statistics*column*quality*grouping:
	
	stream <- unlist(secure.results[(length(g)+1):length(secure.results)])
	names(stream) <- c() 

    # compose a linearized data.frame:

    ns <- length(.g$stat.names)
    nc <- length(x)
    if (.g$.jk.debug) qu <- c("approx. value","approx. protection error","true protection error")
    else qu <- c("approx. value","approx. protection error")
    nq <- length(qu)
    ng <- length(stream)/ns/nc/nq
    .g$return.df <- data.frame(list(
             stat = rep(.g$stat.names,rep(nc*nq*ng,ns))
            ,label = rep(.g$stat.labels,rep(nc*nq*ng,ns))
            ,column = rep(rep(x,rep(nq*ng,nc)),ns)
            ,quality = rep(rep(qu,rep(ng,nq)),ns*nc)
        ))
    for (gbi in .g$group.by)
        .g$return.df[[gbi$name]] <- rep(unlist(lapply(gbi$groups,function(group){group$label})),ns*nc*nq) 
    .g$return.df$result <- stream

	# compose the result structure as a list containing only one mudit structure:
	.g$return.mudit <- 
		list(
			list(
				 type="mudit"
				,dims=c(
					 "statistics"
					,"analysed column"
					,"quality"
					,unlist(lapply(.g$group.by,function(gbi){ gbi$label }))
					)
				,groups=c(
					 list(lapply(.g$stat.labels,function(label){
							list(label=label, level=1)
						}))
					,list(lapply(.g$analyse.labels,function(label){
							list(label=label, level=1)
						}))
					,ifelse(.g$.jk.debug,list(list(
							 list(label="approx. value", level=1)
							,list(label="approx. protection error", level=1)
							,list(label="true protection error", level=1)
						)),list(list(
							 list(label="approx. value", level=1)
							,list(label="approx. protection error", level=1)
						)))
					,lapply(.g$group.by,function(gbi){
						lapply(gbi$groups,function(group){ 
							list(label=group$label, level=group$level) 
						})
					})
				)
				,stream=stream
			)
		)

	if (.g$.jk.debug) {
		print("return.mudit")
		print(.g$return.mudit)
        print("return.df")
        print(.g$return.df)
	}
    return(.g)
}
